from opengen.config.tcp_server_config import TcpServerConfiguration
from opengen.config.ros_config import RosConfiguration
import random
import string
from enum import Enum


class RustAllocator(Enum):
    """Memory Allocator for the auto-generated project"""

    DefaultAllocator = 0
    """Default allocator"""

    JemAlloc = 1
    """Memory allocator: `jemalloc`
    
    Jemalloc is a generic implementation of malloc that emphasises fragmentation
    avoidance
    """

    RpAlloc = 2
    """Memory allocator: `rpmalloc`
    
    Rpmalloc is a very efficient lock-free thread caching 16-byte aligned 
    memory allocator implemented in C.
    """


class BuildConfiguration:
    """Build configuration

    Configuration for the code generator

    """

    #: Debug mode (fast compilation, worse performance)
    DEBUG_MODE = "debug"

    #: Release mode (great performance, very slow compilation)
    RELEASE_MODE = "release"

    def __init__(self, build_dir="."):
        """
        Construct an instance of BuildConfiguration

        :param build_dir: Target directory, defaults to the current directory

        :return: A new instance of BuildConfiguration

        """

        self.__target_system = None
        self.__build_mode = BuildConfiguration.RELEASE_MODE
        self.__rebuild = False
        self.__build_dir = build_dir
        self.__open_version = None
        self.__build_c_bindings = False
        self.__build_python_bindings = False
        self.__ros_config = None
        self.__tcp_interface_config = None
        self.__local_path = None
        self.__allocator = RustAllocator.DefaultAllocator

    # ---------- GETTERS ---------------------------------------------

    @property
    def rebuild(self):
        """Whether to re-build the optimizer from scratch"""
        return self.__rebuild

    @property
    def target_system(self):
        """Target system

        See also: :class:`~opengen.config.build_config.BuildConfiguration.with_target_system`
        """
        return self.__target_system

    @property
    def build_mode(self):
        """
        Build mode (:class:`~opengen.config.build_config.BuildConfiguration.RELEASE_MODE` 
        or :class:`~opengen.config.build_config.BuildConfiguration.DEBUG_MODE`)
        """
        return self.__build_mode

    @property
    def build_dir(self):
        """Directory in which the auto-generated optimizer will be stored"""
        return self.__build_dir

    @property
    def open_version(self):
        """
        OpEn version used with the auto-generated solver

        :return: The method returns either a specific version of OpEn,
            which will be used with the auto-generated optimizer, or `None`,
            in which case, the latest version will be used. You may set your
            preferred version of OpEn with `with_open_version`
        """
        return self.__open_version

    @property
    def local_path(self):
        """Local path of OpEn (if any)"""
        return self.__local_path

    @property
    def build_c_bindings(self):
        """
        Whether to build C bindings 
        """
        return self.__build_c_bindings

    @property
    def build_python_bindings(self):
        """
        Whether to build Python bindings
        """
        return self.__build_python_bindings

    @property
    def tcp_interface_config(self):
        """
        Whether to build a TCP interface
        """
        return self.__tcp_interface_config

    @property
    def ros_config(self) -> RosConfiguration:
        """ROS package configuration

        :return: instance of RosConfiguration
        """
        return self.__ros_config

    @property
    def allocator(self) -> RustAllocator:
        """
        Memory allocator for generated Rust solver
        """
        return self.__allocator

    # ---------- SETTERS ---------------------------------------------

    def with_rebuild(self, do_rebuild):
        """Whether to clean and rebuild the code generator, if it already exists

        :param do_rebuild: if set to True, the target code generator
            will be cleaned and rebuilt from scratch

        :return: current instance of BuildConfiguration
        """
        self.__rebuild = do_rebuild
        return self

    def with_target_system(self, target_system):
        """
        Specify the target system

        :param target_system: target system as string (e.g., use
            "arm-unknown-linux-gnueabihf" or "rpi" for Raspberry Pi).
            Note that you must have installed the target using `rustup`
            if you need to cross-compile.

        :return: current instance of BuildConfiguration
        """
        if target_system.lower() == "rpi":
            self.__target_system = "arm-unknown-linux-gnueabihf"
        else:
            self.__target_system = target_system
        return self

    def with_build_mode(self, build_mode):
        """
        Set the build mode (debug/release)

        :param build_mode: Choose either 'debug' or 'release'; the former is
            fast, but suboptimal, while the latter may take a while to compile,
            but the generated binary is significantly faster

        :return: current instance of BuildConfiguration

        """
        self.__build_mode = build_mode
        return self

    def with_build_directory(self, build_dir):
        """
        Specify the build directory

        :param build_dir: build directory as string

        :return: current instance of BuildConfiguration
        """
        self.__build_dir = build_dir
        return self

    def with_open_version(self, open_version="*", local_path=None):
        """
        Specify the version of OpEn to link to

        :param open_version: version of OpEn (in case you want to
            compile with an older version of OpEn; if not, the latest
            version of OpEn will be used)

        :param local_path: you can compile using a local version
            of OpEn. In that case, you need to provide the full absolute
            path to that local OpEn directory. This option is intended for 
            developers.

        :return: current instance of BuildConfiguration
        """
        self.__open_version = open_version
        self.__local_path = local_path
        return self

    def with_build_c_bindings(self, build_c_bindings=True):
        """
        If activated, OpEn will generate C/C++ bindings for the
        auto-generated solver

        :param build_c_bindings: whether to build C/C++ bindings for
            auto-generated solver; default: `True`, i.e., it suffices
            to call `build_config.with_build_c_bindings()` instead of
            `build_config.with_build_c_bindings(True)`

        :return: current instance of BuildConfiguration
        """
        self.__build_c_bindings = build_c_bindings
        return self

    def with_build_python_bindings(self, build_python_bindings=True):
        """
        If activated, OpEn will generate python bindings for the
        auto-generated solver

        :param build_python_bindings: whether to build python bindings for
            auto-generated solver; default: `True`, i.e., it suffices
            to call `build_config.with_build_python_bindings()` instead of
            `build_config.with_build_python_bindings(True)`

        :return: current instance of BuildConfiguration
        """
        self.__build_python_bindings = build_python_bindings
        return self

    def with_ros(self, ros_config: RosConfiguration):
        """
        Activates the generation of a ROS package. The caller must provide an
        instance of RosConfiguration

        :param ros_config: Configuation of ROS package

        :return: current instance of BuildConfiguration
        """
        self.__build_c_bindings = True  # no C++ bindings, no ROS package mate
        self.__ros_config = ros_config
        return self

    def with_tcp_interface_config(self, tcp_interface_config=TcpServerConfiguration()):
        """
        Specify a TCP server configuration object

        :param tcp_interface_config: Custom TCP server configuration

        :return: current instance of BuildConfiguration
        """
        self.__tcp_interface_config = tcp_interface_config
        return self

    def with_allocator(self, allocator: RustAllocator):
        """Specify a Rust memory allocator. 

        With this method the user can choose an alternative memory allocator
        such as `jemalloc` and `rpalloc`. For example, if you choose `jemalloc`
        as your memory allocator, the autogenerated project will have a 
        `Cargo.toml` file where `optimization-engine` is loaded as a dependency
        with the "jem" feature.

        :param allocator: allocator; instance of `RustAllocator`

        :return: current instance of BuildConfiguration
        """
        self.__allocator = allocator
        return self

    def to_dict(self):
        build_dict = {
            "target_system": self.__target_system,
            "build_mode": self.__build_mode,
            "rebuild": self.__rebuild,
            "build_directory": self.__build_dir,
            "open_version": self.__open_version,
            "build_c_bindings": self.__build_c_bindings,
            "build_python_bindings": self.__build_python_bindings,
        }
        if self.__tcp_interface_config is not None:
            build_dict["tcp_interface_config"] = self.__tcp_interface_config.to_dict()
        if self.__ros_config is not None:
            build_dict["ros_config"] = self.__ros_config.to_dict()
        return build_dict



from .constraint import Constraint
import sys


class Halfspace(Constraint):
    """Halfspace constraint

    A halfspace is a set of the form :math:`H = \{c^\intercal x \leq b\}`, where `c` is a given
    vector and `b` is a constant scalar.
    """

    def __init__(self, normal_vector, offset: float):
        """Construct a new halfspace :math:`H = \{c^\intercal x \leq b\}`

        :param normal_vector: vector `c`

        :param offset: parameter `b`
        """
        self.__normal_vector = [float(x) for x in normal_vector]
        self.__offset = float(offset)

    @property
    def normal_vector(self):
        """Vector `c`
        """
        return self.__normal_vector

    @property
    def offset(self):
        """Parameter `b`
        """
        return self.__offset

    def dimension(self):
        """
        Dimension of the halfspace

        :return: length/dimension of normal vector
        """
        return len(self.__normal_vector)

    def project(self, u):
        raise NotImplementedError()

    def distance_squared(self, u):
        raise NotImplementedError()

    def is_convex(self):
        """
        A halfspace is a convex set

        :return: ``True``
        """
        return True

    def is_compact(self):
        """Whether the set is compact

        H is compact iff :math:`b < 0` and :math:`c = 0`, in which case `H` is empty
        """
        eps = sys.float_info.epsilon
        # if b < 0 and c = 0, then H is empty, hence compact
        if self.offset < 0 and sum([self.normal_vector[idx]
                                    for idx in range(self.dimension())]) < eps:
            return True
        return False



import casadi.casadi as cs
import numpy as np
from .constraint import Constraint
import opengen.functions as fn


class Zero(Constraint):
    """A set that contains only the origin

    The singleton :math:`\{0\}`

    """

    def __init__(self):
        """
        Constructor for set :math:`Z = \{0\}`

        """

    def distance_squared(self, u):
        return fn.norm2_squared(u)

    def project(self, u):
        raise NotImplementedError()

    def is_convex(self):
        return True

    def is_compact(self):
        return True



class TcpServerConfiguration:
    """TCP server configuration"""

    def __init__(self, bind_ip='127.0.0.1', bind_port=8333):
        """Configuration of the TCP server

        :param bind_ip: IP address of generated TCP server. The default
            value is "127.0.0.1" (localhost). Use "0.0.0.0" for the generated
            TCP server to bind on all IPs.

        :param bind_port: Port on which the generated TCP server will bind.
            The default is 8333. Make sure you use an available port and avoid using
            privileged ports (i.e., 1 to 1024), well-known ports that are potentially
            used by other services and you should also avoid ephemeral ports
            (32768 to 65535 on Linux, 1025 to 5000 on Windows)

        :returns: new instance of TcpServerConfiguration, which can then be
            provided to an instance of `OpEnOptimizerBuilder` via `enable_tcp_interface`
        """
        self.__bind_ip = bind_ip
        self.__bind_port = bind_port

    @property
    def bind_ip(self):
        """IP at which the TCP server should bind, as a string

        :return: TCP server IP
        """
        return self.__bind_ip

    @property
    def bind_port(self):
        """Port at which the TCP server should bind, as int

        :return: TCP server port
        """
        return self.__bind_port

    def to_dict(self):
        return {
            "ip": self.__bind_ip,
            "port": self.__bind_port
        }



from .solver_status import SolverStatus
from .solver_error import SolverError


class SolverResponse:
    """Stores a solver response of type SolverStatus or SolverError."""

    def __init__(self, d):
        """Constructs instance of :class:`~opengen.tcp.solver_response.SolverResponse`

        This constructor is not to be used by the end-user

        :param d: dictionary containing either status or error attributes
        :type d: dictionary

        :return: New instance of :class:`~opengen.tcp.solver_response.SolverResponse`
        """
        if 'Error' in d.values():
            self.__response = SolverError(d)
        else:
            self.__response = SolverStatus(d)

    def is_ok(self):
        """Determines if response is OK.

        This method should always be called first when obtaining a server response.
        """
        return isinstance(self.__response, SolverStatus)

    def get(self):
        """
        Returns response, which is an instance of SolverStatus,
        if the call was successful, or an instance of SolverError
        otherwise. It is recommended that you use 
        :class:`~opengen.tcp.solver_response.SolverResponse.is_ok` to check
        whether the call has succeeded first
        """
        return self.__response

    def __getitem__(self, key):
        return getattr(self.__response, key)



class SolverError:
    """Class for storing solver status in the event of an error."""

    def __init__(self, error):
        """Constructs instance of :class:`~opengen.tcp.solver_error.SolverError`

        :param error: dictionary containing error attributes

        :return: New instance of :class:`~opengen.tcp.solver_error.SolverError`
        """

        for k, v in error.items():
            attribute_name = "__{}".format(k)
            setattr(self, attribute_name, v)

    @property
    def code(self):
        """Error code

        Possible error codes are:

        - **1000**: Invalid request: Malformed or invalid JSON
        - **1600**: Initial guess has incomplete dimensions
        - **1700**: Wrong dimension of Lagrange multipliers
        - **2000**: Problem solution failed (solver error)
        - **3003**: Parameter vector has wrong length

        :return: Error code
        """
        return self.__dict__["__code"]

    @property
    def message(self):
        """
        Returns an appropriate error message matching the error code

        :return: Error message
        :rtype: str
        """
        return self.__dict__["__message"]



class SolverConfiguration:
    """Configuration of solver parameters

    """

    def __init__(self):
        """Construct an instance of solver configuration parameters

        Returns:
            :return: New instance of SolverConfiguration

        """
        self.__tolerance = 1e-4
        self.__initial_tolerance = 1e-4
        self.__lbfgs_memory = 10
        self.__max_inner_iterations = 500
        self.__max_outer_iterations = 10
        self.__constraints_tolerance = 1e-4
        # For the initial penalty, None means that the actual value will be computed
        # in Rust (see templates/optimizer.rs)
        self.__initial_penalty = None
        self.__penalty_weight_update_factor = 5.0
        self.__max_duration_micros = 5000000
        self.__inner_tolerance_update_factor = 0.1
        self.__sufficient_decrease_coefficient = 0.1
        self.__cbfgs_alpha = None
        self.__cbfgs_epsilon = None
        self.__cbfgs_sy_epsilon = None
        self.__do_preconditioning = False  # alpha version of preconditioning: optional

    # --------- GETTERS -----------------------------

    @property
    def sufficient_decrease_coefficient(self):
        """Sufficient decrease coefficient"""
        return self.__sufficient_decrease_coefficient

    @property
    def initial_penalty(self):
        """Initial penalty"""
        return self.__initial_penalty

    @property
    def cbfgs_alpha(self):
        return self.__cbfgs_alpha

    @property
    def cbfgs_epsilon(self):
        return self.__cbfgs_epsilon

    @property
    def cbfgs_sy_epsilon(self):
        return self.__cbfgs_sy_epsilon

    @property
    def tolerance(self):
        """Tolerance of inner solver"""
        return self.__tolerance

    @property
    def initial_tolerance(self):
        """Initial tolerance of inner solver"""
        return self.__initial_tolerance

    @property
    def inner_tolerance_update_factor(self):
        """"Update factor for inner tolerance"""
        return self.__inner_tolerance_update_factor

    @property
    def lbfgs_memory(self):
        """LBFGS memory for the inner solver"""
        return self.__lbfgs_memory

    @property
    def max_inner_iterations(self):
        """Maximum number of iterations for the inner solver"""
        return self.__max_inner_iterations

    @property
    def constraints_tolerance(self):
        """Tolerance on the satisfaction of the constraints"""
        return self.__constraints_tolerance

    @property
    def max_outer_iterations(self):
        """Maximum number of iterations for the outer solver"""
        return self.__max_outer_iterations

    @property
    def penalty_weight_update_factor(self):
        """Multiplicative factor for the update of the penalty weights"""
        return self.__penalty_weight_update_factor

    @property
    def max_duration_micros(self):
        """Maximum execution time in microseconds

        Returns:
            :return: Integer value
        """
        return self.__max_duration_micros

    @property
    def preconditioning(self):
        """Whether an automatic preconditioning should be applied

        Returns:        
            :return: True iff preconditioning is active
        """
        return self.__do_preconditioning

    # --------- SETTERS -----------------------------

    def with_sufficient_decrease_coefficient(self, sufficient_decrease_coefficient):
        """
        Specify the sufficient decrease coefficient of the algorithm

            :param sufficient_decrease_coefficient: sufficient decrease coefficient

            :returns: The current object
        """
        if sufficient_decrease_coefficient <= 0.0 or sufficient_decrease_coefficient >= 1.0:
            raise Exception("sufficient decrease coeff must be in (0,1)")
        self.__sufficient_decrease_coefficient = sufficient_decrease_coefficient
        return self

    def with_initial_penalty(self, initial_penalty):
        """Initial penalty

        If preconditioning is activated, then the initial penalty is computed internally
        following the recommendations of the book of Brigin and Martinez (Chapter 12).
        If you enable the preconditioning and you use this method, then you will be
        overriding the value of the initial penalty.

        If preconditioning is not enabled, you can set the initial penalty using this
        method; if you don't do so, the solver will use the default value (which is 1.0).

        :param initial_penalty: initial value of penalty

        :returns: The current object
        """
        if initial_penalty is None:
            self.__initial_penalty = None
            return

        if initial_penalty <= 0:
            raise Exception("Initial penalty must be >0")

        self.__initial_penalty = float(initial_penalty)
        return self

    def with_tolerance(self, tolerance):
        """Specify tolerance

        :param tolerance: tolerance

        :raises: Exception: if tolerance <= 0

        :returns: The current object
        """
        if tolerance <= 0:
            raise Exception("The tolerance must be >0")
        self.__tolerance = float(tolerance)
        return self

    def with_initial_tolerance(self, initial_tolerance):
        """
        Specify the initial tolerance

        :param initial_tolerance: initial tolerance

        :returns: The current object
        """
        if initial_tolerance <= 0:
            raise Exception("The initial tolerance must be >0")
        self.__initial_tolerance = float(initial_tolerance)
        return self

    def with_inner_tolerance_update_factor(self, inner_tol_update_factor):
        """Tolerance update factor

        The tolerance is initially given by 
        :py:meth:`~opengen.config.solver_config.SolverConfiguration.with_initial_tolerance`
        and it is then updated by this update factor until the target tolerance
        which is given by 
        :py:meth:`~opengen.config.solver_config.SolverConfiguration.with_tolerance`
        """
        if inner_tol_update_factor <= 0 or inner_tol_update_factor > 1:
            raise Exception("The tolerance update factor must be in [0, 1)")

        self.__inner_tolerance_update_factor = float(inner_tol_update_factor)
        return self

    def with_lbfgs_memory(self, lbfgs_memory):
        """Specify L-BFGS memory

        :param lbfgs_memory: LBFGS memory

        :raises: It is required that the L-BFGS memory is larger than or
            equal to 2, otherwise an Exception is raised

        :return: Returns the current instance of SolverConfiguration
        """
        if lbfgs_memory < 2:
            raise Exception("The L-BFGS memory must be at least equal to 2")
        self.__lbfgs_memory = int(lbfgs_memory)
        return self

    def with_max_inner_iterations(self, max_iters):
        """Maximum number of inner iterations

        :param max_iters: maximum number of iterations

        :return: The current object
        """
        if max_iters < 1:
            raise Exception(
                "The maximum number of inner iterations must be at least equal to 1")
        self.__max_inner_iterations = int(max_iters)
        return self

    def with_delta_tolerance(self, constraints_tolerance):
        """Tolerance on constraint violation

        :param constraints_tolerance: tolerance delta (related to constraint violation)

        :return: the current object
        """
        if constraints_tolerance <= 0:
            raise Exception(
                "The constraints tolerance must be strictly positive")
        self.__constraints_tolerance = float(constraints_tolerance)
        return self

    def with_max_outer_iterations(self, max_outer_iterations):
        """Maximum number of outer iterations

        :param max_outer_iterations: maximum number of outer iterations

        :return: the current object
        """
        if max_outer_iterations < 1:
            raise Exception(
                "The maximum number of outer iterations must be at least equal to 1")
        self.__max_outer_iterations = int(max_outer_iterations)
        return self

    def with_penalty_weight_update_factor(self, penalty_weight_update_factor):
        """Penalty update factor

        At every outer iteration of the penalty method, the weights are
        multiplied by this factor.

        :param penalty_weight_update_factor: penalty weight update factor

        :raises: Exception, if the update factor is less than 1.0

        :return: the current object
        """
        if penalty_weight_update_factor < 1.0:
            raise Exception("The penalty update factor needs to be >= 1")
        self.__penalty_weight_update_factor = float(
            penalty_weight_update_factor)
        return self

    def with_max_duration_micros(self, max_duration_micros):
        """Specify the maximum duration in microseconds (must be an integer)

        The solver will interrupt the computation after this time limit and 
        will return the current iterate.

        :param max_duration_micros: maximum execution duration in microseconds (integer)

        :raises: Exception: if `max_duration_micros` is less than 1

        :return: The current object
        """
        if max_duration_micros < 1:
            raise Exception(
                "The maximum duration (in microseconds) must be >= 1")
        self.__max_duration_micros = int(max_duration_micros)
        return self

    def with_cbfgs_parameters(self, alpha, epsilon, sy_epsilon):
        """Specify the CBFGS parameters alpha and epsilon

        :param alpha: CBFGS parameter alpha
        :param epsilon: CBFGS parameter epsilon
        :param sy_epsilon: Tolerance on the s-y inner product

        :returns: the current object
        """
        if epsilon < 0.0:
            raise Exception("CBFGS parameter epsilon must be positive")

        if alpha < 0.0:
            raise Exception("CBFGS parameter alpha must be positive")

        self.__cbfgs_epsilon = epsilon
        self.__cbfgs_alpha = alpha
        self.__cbfgs_sy_epsilon = sy_epsilon
        return self

    def with_preconditioning(self, do_preconditioning=True):
        """Whether to apply preconditioning using the approach of [1]

        Note that this overrides the computation of the initial penalty

        Note also that unless this method is called, no preconditioning is
        applied (this may change in a future release; we may make enable
        preconditioning by default)

        [1] E.G. Birgin and J.M. Martinez, Practical Augmented Lagrangian
        Methods for Constrained Optimization, SIAM, 2014

        :param do_preconditioning: whether to precondition

        :returns: the current object
        """
        self.__do_preconditioning = do_preconditioning
        return self

    def to_dict(self):
        return {
            "tolerance": self.__tolerance,
            "initial_tolerance": self.__initial_tolerance,
            "lbfgs_memory": self.__lbfgs_memory,
            "max_inner_iterations": self.__max_inner_iterations,
            "max_outer_iterations": self.__max_outer_iterations,
            "constraints_tolerance": self.__constraints_tolerance,
            "initial_penalty": self.__initial_penalty,
            "penalty_weight_update_factor": self.__penalty_weight_update_factor,
            "max_duration_micros": self.__max_duration_micros,
            "inner_tolerance_update_factor": self.__inner_tolerance_update_factor,
            "sufficient_decrease_coefficient": self.__sufficient_decrease_coefficient,
            "cbfgs_alpha": self.__cbfgs_alpha,
            "cbfgs_epsilon": self.__cbfgs_epsilon,
            "cbfgs_sy_epsilon": self.__cbfgs_sy_epsilon,
            "do_preconditioning": self.__do_preconditioning
        }



import casadi.casadi as cs
import opengen as og
import os
import numpy as np

TEST_DIR = ".python_test_build/benchmarkable"


def get_open_local_absolute_path():
    cwd = os.getcwd()
    return cwd.split('open-codegen')[0]


def solver_configuration(do_precond=False):
    solver_config = og.config.SolverConfiguration() \
        .with_tolerance(1e-6) \
        .with_initial_tolerance(1e-4) \
        .with_delta_tolerance(1e-5) \
        .with_max_inner_iterations(1000) \
        .with_preconditioning(do_precond)
    return solver_config


def build_configuration():
    return og.config.BuildConfiguration() \
        .with_open_version(local_path=get_open_local_absolute_path()) \
        .with_build_mode(og.config.BuildConfiguration.RELEASE_MODE) \
        .with_build_directory(TEST_DIR) \
        .with_build_python_bindings()


def benchmark1(optimizer_name, do_precond=False):
    u = cs.MX.sym("u", 5)
    p = cs.MX.sym("p", 3)
    f2 = u[0] - p[2]
    phi = og.functions.rosenbrock(u, cs.vertcat(p[0], p[1]))  # cost function
    bounds = og.constraints.Ball2(None, 1.5)
    meta = og.config.OptimizerMeta() \
        .with_optimizer_name(optimizer_name)
    problem = og.builder.Problem(u, p, phi) \
        .with_penalty_constraints(f2) \
        .with_constraints(bounds)
    build_config = build_configuration()
    solver_config = solver_configuration(do_precond)
    og.builder.OpEnOptimizerBuilder(problem, meta, build_config, solver_config).build()


def navigation_problem(with_obstacles=False):
    (nu, nx, N, L, ts) = (2, 3, 60, 0.5, 0.1)
    (xref, yref, thetaref) = (2, 2, 0.2)
    (x_obs, y_obs, r_obs) = (0, 0, 1)
    (q, qtheta, r, qN, qthetaN) = (10, 0.1, 1, 200, 2)

    u = cs.SX.sym('u', nu * N)
    z0 = cs.SX.sym('z0', nx)

    (x, y, theta) = (z0[0], z0[1], z0[2])
    cost = 0
    obstacle_avoidance = []
    for t in range(0, nu * N, nu):
        cost += q * ((x - xref) ** 2 + (y - yref) ** 2) + qtheta * (theta - thetaref) ** 2
        u_t = u[t:t + 2]
        theta_dot = (u_t[1] * cs.cos(theta) - u_t[0] * cs.sin(theta)) / L
        cost += r * cs.dot(u_t, u_t)
        x += ts * (u_t[0] + L * cs.sin(theta) * theta_dot)
        y += ts * (u_t[1] - L * cs.cos(theta) * theta_dot)
        theta += ts * theta_dot
        obs = 1000 * cs.fmax(0, r_obs**2 - (x-x_obs) ** 2 - (y-y_obs)**2)
        obstacle_avoidance = cs.vertcat(obstacle_avoidance, obs)

    cost += qN * ((x - xref) ** 2 + (y - yref) ** 2) + qthetaN * (theta - thetaref) ** 2

    umin = [-3.0] * (nu * N)
    umax = [3.0] * (nu * N)
    bounds = og.constraints.Rectangle(umin, umax)

    problem = og.builder.Problem(u, z0, cost).with_constraints(bounds)
    if with_obstacles:
        problem = problem.with_penalty_constraints(obstacle_avoidance)
    return problem


def benchmark2(optimizer_name, with_obstacles=False, do_precond=False):
    """Navigation"""
    problem = navigation_problem(with_obstacles)
    build_config = build_configuration()
    meta = og.config.OptimizerMeta() \
        .with_optimizer_name(optimizer_name)
    solver_config = og.config.SolverConfiguration() \
        .with_tolerance(1e-5) \
        .with_penalty_weight_update_factor(2) \
        .with_max_inner_iterations(1000) \
        .with_preconditioning(do_precond)
    builder = og.builder.OpEnOptimizerBuilder(problem,
                                              meta,
                                              build_config,
                                              solver_config)
    builder.build()


benchmark1("benchmark1", do_precond=False)
benchmark1("benchmark1p", do_precond=True)
benchmark2("benchmark2", do_precond=False)
benchmark2("benchmark2p", do_precond=True)
benchmark2("benchmark2o", with_obstacles=True, do_precond=False)
benchmark2("benchmark2op", with_obstacles=True, do_precond=True)


import time
import casadi.casadi as cs
import opengen as og
import os
import sys
import numpy as np
from importlib import import_module

ITERS = 10
ROUNDS = 1500

#
# Before you run this, run
# python prepare_benchmarks.py
#
# Run benchmark with:
# py.test test/benchmark_open.py --benchmark-warmup-iterations=20
#
# To generate a histogram, do:
# py.test test/benchmark_open.py --benchmark-warmup-iterations=20 --benchmark-histogram=out
# qlmanage -t -s 1000 -o . out.svg
#

TEST_DIR = ".python_test_build/benchmarkable"


for name in os.listdir(TEST_DIR):
    sys.path.insert(1, os.path.join(TEST_DIR, name))
    exec(f"{name} = import_module(\"{name}\")")
    exec(f"solver_{name} = {name}.solver()")


def get_open_local_absolute_path():
    cwd = os.getcwd()
    return cwd.split('open-codegen')[0]


def t_benchmark1(solver):
    a = np.random.uniform(0.5, 2)
    b = np.random.uniform(0.5, 15)
    c = np.random.uniform(0.9, 3)
    _sol = solver.run([a, b, c])


def t_benchmark2(solver):
    x0 = np.random.uniform(-3.5, -2)
    y0 = np.random.uniform(-2.5, 2.5)
    # th0 = np.random.uniform(-0.3, 0.3)
    _sol = solver.run([x0, y0, 0])


def test_benchmark1(benchmark):
    benchmark.pedantic(t_benchmark1, kwargs={'solver': solver_benchmark1},
                       iterations=ITERS, rounds=ROUNDS)


def test_benchmark1p(benchmark):
    benchmark.pedantic(t_benchmark1, kwargs={'solver': solver_benchmark1p},
                       iterations=ITERS, rounds=ROUNDS)


def test_benchmark2(benchmark):
    benchmark.pedantic(t_benchmark2, kwargs={'solver': solver_benchmark2},
                       iterations=ITERS, rounds=ROUNDS)


def test_benchmark2p(benchmark):
    benchmark.pedantic(t_benchmark2, kwargs={'solver': solver_benchmark2p},
                       iterations=ITERS, rounds=ROUNDS)


def test_benchmark2o(benchmark):
    benchmark.pedantic(t_benchmark2, kwargs={'solver': solver_benchmark2o},
                       iterations=ITERS, rounds=ROUNDS)


def test_benchmark2op(benchmark):
    benchmark.pedantic(t_benchmark2, kwargs={'solver': solver_benchmark2op},
                       iterations=ITERS, rounds=ROUNDS)
